/**************************************************************************************************
  Filename:       modbusCentral.c
  Revised:        $Date: 2011-06-20 11:57:59 -0700 (Mon, 20 Jun 2011) $
  Revision:       $Revision: 28 $

  Description:    This file contains the Simple BLE Central sample application 
                  for use with the CC2540 Bluetooth Low Energy Protocol Stack.

  Copyright 2010 Texas Instruments Incorporated. All rights reserved.

  IMPORTANT: Your use of this Software is limited to those specific rights
  granted under the terms of a software license agreement between the user
  who downloaded the software, his/her employer (which must be your employer)
  and Texas Instruments Incorporated (the "License").  You may not use this
  Software unless you agree to abide by the terms of the License. The License
  limits your use, and you acknowledge, that the Software may not be modified,
  copied or distributed unless embedded on a Texas Instruments microcontroller
  or used solely and exclusively in conjunction with a Texas Instruments radio
  frequency transceiver, which is integrated into your product.  Other than for
  the foregoing purpose, you may not use, reproduce, copy, prepare derivative
  works of, modify, distribute, perform, display or sell this Software and/or
  its documentation for any purpose.

  YOU FURTHER ACKNOWLEDGE AND AGREE THAT THE SOFTWARE AND DOCUMENTATION ARE
  PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
  INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY, TITLE,
  NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL
  TEXAS INSTRUMENTS OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER CONTRACT,
  NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR OTHER
  LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
  INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE
  OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT
  OF SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
  (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

  Should you have any questions regarding your right to use this Software,
  contact Texas Instruments Incorporated at www.TI.com.
**************************************************************************************************/

/*********************************************************************
 * INCLUDES
 */

#include "bcomdef.h"
#include "OSAL.h"
#include "OSAL_PwrMgr.h"
#include "OnBoard.h"
#include "hal_led.h"
#include "hal_key.h"
#include "hal_lcd.h"
#include "hal_uart.h"
#include "gatt.h"
#include "ll.h"
#include "hci.h"
#include "gapgattserver.h"
#include "gattservapp.h"
#include "central.h"
#include "gapbondmgr.h"
#include "SerialProfile.h"
#include "modbusCentral.h"
#include "modbusDataTransfer.h"   

uint16 tull;
/*********************************************************************
 * MACROS
 */

// Length of bd addr as a string
#define B_ADDR_STR_LEN                        15

/*********************************************************************
 * CONSTANTS
 */
//serialprofile UUIDs
CONST uint8 dataInUUID[ATT_UUID_SIZE] = {TI_UUID(SERIALPROFILE_DATA_IN_UUID)};
CONST uint8 dataOutUUID[ATT_UUID_SIZE] = {TI_UUID(SERIALPROFILE_DATA_OUT_UUID)};

// Maximum number of scan responses
#define DEFAULT_MAX_SCAN_RES                  8

// LED blink periods
#define MBC_IDLE_BLINK_PERIOD                 1000  
#define MBC_DISC_BLINK_PERIOD                 400
   
// Key press debounce time
#define MBC_KEY_DEBOUNCE                      50
   
// Scan duration in ms
#define DEFAULT_SCAN_DURATION                 20000

// Discovey mode (limited, general, all)
#define DEFAULT_DISCOVERY_MODE                DEVDISC_MODE_ALL

// TRUE to use active scan
#define DEFAULT_DISCOVERY_ACTIVE_SCAN         TRUE

// TRUE to use white list during discovery
#define DEFAULT_DISCOVERY_WHITE_LIST          FALSE

// TRUE to use high scan duty cycle when creating link
#define DEFAULT_LINK_HIGH_DUTY_CYCLE          FALSE

// TRUE to use white list when creating link
#define DEFAULT_LINK_WHITE_LIST               FALSE

// Slave latency to use if automatic parameter update request is enabled
#define DEFAULT_UPDATE_SLAVE_LATENCY          0

// Default passcode
#define DEFAULT_PASSCODE                      19655

// Default GAP pairing mode
#define DEFAULT_PAIRING_MODE                  GAPBOND_PAIRING_MODE_WAIT_FOR_REQ

// Default MITM mode (TRUE to require passcode or OOB when pairing)
#define DEFAULT_MITM_MODE                     FALSE

// Default bonding mode, TRUE to bond
#define DEFAULT_BONDING_MODE                  TRUE

// Default GAP bonding I/O capabilities
#define DEFAULT_IO_CAPABILITIES               GAPBOND_IO_CAP_DISPLAY_ONLY

// Default service discovery timer delay in ms
#define DEFAULT_SVC_DISCOVERY_DELAY           1000

// TRUE to filter discovery results on desired service UUID
#define DEFAULT_DEV_DISC_BY_SVC_UUID          TRUE

// Application states
enum
{
  BLE_STATE_IDLE,
  BLE_STATE_CONNECTING,
  BLE_STATE_CONNECTED,
  BLE_STATE_DISCONNECTING
};

// Discovery states
enum
{
  BLE_DISC_STATE_IDLE,                // Idle
  BLE_DISC_STATE_SVC,                 // Service discovery
  BLE_DISC_STATE_CHAR                 // Characteristic discovery
};

/*********************************************************************
 * TYPEDEFS
 */

typedef struct
{
  uint16 declHandle;
  uint8 properties;
  uint16 handle;
  uint8 uuid[ATT_UUID_SIZE];
} handleValue128bit_t;

/*********************************************************************
 * GLOBAL VARIABLES
 */
bool notificationsEnabled = false;
/*********************************************************************
 * EXTERNAL VARIABLES
 */

/*********************************************************************
 * EXTERNAL FUNCTIONS
 */

/*********************************************************************
 * LOCAL VARIABLES
 */

// Task ID for internal task/event processing
static uint8 modbusTaskId;

// GAP GATT Attributes
static const uint8 modbusDeviceName[GAP_DEVICE_NAME_LEN] = "Modbus Central";

// Scanning state
static uint8 modbusScanning = FALSE;

// Connection handle of current connection 
static uint16 modbusConnHandle = GAP_CONNHANDLE_INIT;

// Application state
static uint8 modbusState = BLE_STATE_IDLE;

// Discovery state
static uint8 modbusDiscState = BLE_DISC_STATE_IDLE;

// Discovered service start and end handle
static uint16 serialSvcStartHdl = 0;
static uint16 serialSvcEndHdl = 0;

// Discovered characteristic handle
static uint16 dataInHandle = 0;
static uint16 dataOutHandle = 0;//0x28; //read from device monitor

// GATT read/write procedure state
//static bool procedureInProgress = FALSE;

// Message buffer
uint8 modbusMessage[256];
uint16 modbusCurrentLength = 0;
/*********************************************************************
 * LOCAL FUNCTIONS
 */
static void modbusCentralProcessGATTMsg( gattMsgEvent_t *pMsg );
static void modbusCentralEventCB( gapCentralRoleEvent_t *pEvent );
static void modbusCentral_HandleKeys( uint8 shift, uint8 keys );
static void modbusCentral_ProcessOSALMsg( osal_event_hdr_t *pMsg );
static void modbusGATTDiscoveryEvent( gattMsgEvent_t *pMsg );
static void modbusCentralStartDiscovery( void );
static bool modbusFindSvcUuid( uint16 uuid, uint8 *pData, uint8 dataLen );

/*********************************************************************
 * PROFILE CALLBACKS
 */

// GAP Role Callbacks
static const gapCentralRoleCB_t modbusRoleCB =
{
  NULL,//modbusCentralRssiCB,       // RSSI callback
  modbusCentralEventCB       // Event callback
};

// Bond Manager Callbacks
static const gapBondCBs_t modbusBondCB =
{
  NULL,//modbusCentralPasscodeCB,
  NULL//modbusCentralPairStateCB
};

/*********************************************************************
 * PUBLIC FUNCTIONS
 */

/*****************************************************************//**
 * @fn      ModbusCentral_Init
 *
 * @brief   Initialization function for the Modbus Central App Task.
 *          This is called during initialization and should contain
 *          any application specific initialization (ie. hardware
 *          initialization/setup, table initialization, power up
 *          notification).
 *
 * @param   task_id - the ID assigned by OSAL.  This ID should be
 *                    used to send messages and set timers.
 *
 * @return  none
 */
void ModbusCentral_Init( uint8 task_id )
{
  modbusTaskId = task_id;

  // Setup Central Profile
  {
    uint8 scanRes = DEFAULT_MAX_SCAN_RES;
    GAPCentralRole_SetParameter ( GAPCENTRALROLE_MAX_SCAN_RES, sizeof( uint8 ), &scanRes );
  }
  
  // Setup GAP
  GAP_SetParamValue( TGAP_GEN_DISC_SCAN, DEFAULT_SCAN_DURATION );
  GAP_SetParamValue( TGAP_LIM_DISC_SCAN, DEFAULT_SCAN_DURATION );
  GGS_SetParameter( GGS_DEVICE_NAME_ATT, GAP_DEVICE_NAME_LEN, (uint8 *) modbusDeviceName );

  // Setup the GAP Bond Manager
  {
    uint32 passkey = DEFAULT_PASSCODE;
    uint8 pairMode = DEFAULT_PAIRING_MODE;
    uint8 mitm = DEFAULT_MITM_MODE;
    uint8 ioCap = DEFAULT_IO_CAPABILITIES;
    uint8 bonding = DEFAULT_BONDING_MODE;
    GAPBondMgr_SetParameter( GAPBOND_DEFAULT_PASSCODE, sizeof( uint32 ), &passkey );
    GAPBondMgr_SetParameter( GAPBOND_PAIRING_MODE, sizeof( uint8 ), &pairMode );
    GAPBondMgr_SetParameter( GAPBOND_MITM_PROTECTION, sizeof( uint8 ), &mitm );
    GAPBondMgr_SetParameter( GAPBOND_IO_CAPABILITIES, sizeof( uint8 ), &ioCap );
    GAPBondMgr_SetParameter( GAPBOND_BONDING_ENABLED, sizeof( uint8 ), &bonding );
  }  

  // Initialize GATT Client
  VOID GATT_InitClient();

  // Register to receive incoming ATT Indications/Notifications
  GATT_RegisterForInd( modbusTaskId );

  // Initialize GATT attributes
  GGS_AddService( GATT_ALL_SERVICES );         // GAP
  GATTServApp_AddService( GATT_ALL_SERVICES ); // GATT attributes

  // Register for all key events - This app will handle all key events
  RegisterForKeys( modbusTaskId );
  
  // makes sure LEDs are off
  HalLedSet( (HAL_LED_1 | HAL_LED_2), HAL_LED_MODE_OFF );
  
  // Disable clock divide on halt
  HCI_EXT_ClkDivOnHaltCmd( HCI_EXT_DISABLE_CLK_DIVIDE_ON_HALT ); //changed from enable
  HCI_EXT_HaltDuringRfCmd( HCI_EXT_HALT_DURING_RF_DISABLE ); //added to keep the processor from shutting down

  
  Modbus_initUART();
  Modbus_initTimer();
#if defined CC2540_RS485
  //RS485_InitBoard();
#endif
  
  LED_BLINK_START();
  
  // Setup a delayed profile startup
  osal_set_event( modbusTaskId, START_DEVICE_EVT );
}

/*****************************************************************//**
 * @fn      ModbusCentral_ProcessEvent
 *
 * @brief   Modbus Central Application Task event processor.  This function
 *          is called to process all events for the task.  Events
 *          include timers, messages and any other user defined events.
 *
 * @param   task_id  - The OSAL assigned task ID.
 * @param   events - events to process.  This is a bit map and can
 *                   contain more than one event.
 *
 * @return  events not processed
 */
uint16 ModbusCentral_ProcessEvent( uint8 task_id, uint16 events )
{
  
  VOID task_id; // OSAL required parameter that isn't used in this function
  
  if ( events & SYS_EVENT_MSG )
  {
    uint8 *pMsg;

    if ( (pMsg = osal_msg_receive( modbusTaskId )) != NULL )
    {
      modbusCentral_ProcessOSALMsg( (osal_event_hdr_t *)pMsg );

      // Release the OSAL message
      VOID osal_msg_deallocate( pMsg );
    }

    // return unprocessed events
    return (events ^ SYS_EVENT_MSG);
  }

  if ( events & START_DEVICE_EVT )
  {
    // Start the Device
    VOID GAPCentralRole_StartDevice( (gapCentralRoleCB_t *) &modbusRoleCB );

    // Register with bond manager after starting device
    GAPBondMgr_Register( (gapBondCBs_t *) &modbusBondCB );

    RS485_InitBoard(modbusTaskId); //initialize the peripherals
    RS485_enableTest();
    
    
    return ( events ^ START_DEVICE_EVT );
  }
  
   if (events & MBC_FINAL_FRAGMENT_EVT ){
    //send the final fragments
    uint8 data[CHARACTERISTIC_LEN];
    uint8 tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    
    CLEAR_TEST();
    while (tag >= CHARACTERISTIC_LEN){
      Modbus_GetFragment(data, CHARACTERISTIC_LEN-1);
      Modbus_WriteDataIn(data, CHARACTERISTIC_LEN);
      tag -= (CHARACTERISTIC_LEN-1);
      //tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    }
    //tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    HalUARTRead(SDT_UART_PORT, data+1, tag);
    data[0] = tag;
    if (tag != 0){
      Modbus_WriteDataIn(data, tag+1);
    }
    return ( events & ~(MBC_FRAGMENT_READY_EVT | MBC_FINAL_FRAGMENT_EVT) );    
  }
  
  if ( events & MBC_FRAGMENT_READY_EVT ) {
    //send a fragment
    uint8 data[CHARACTERISTIC_LEN];
    Modbus_GetFragment(data, CHARACTERISTIC_LEN-1);
    if (data != 0){
      Modbus_WriteDataIn(data, CHARACTERISTIC_LEN);
    }      
    return ( events ^ MBC_FRAGMENT_READY_EVT );
  } 

  if ( events & START_DISCOVERY_EVT )
  {
    modbusCentralStartDiscovery( );
    
    return ( events ^ START_DISCOVERY_EVT );
  }
  
  if ( events & MBC_KEY_EVT) 
  {
    //handle keys
    modbusCentral_HandleKeys(0,0);
    
    return (events ^ MBC_KEY_EVT);
  }
  
  if ( events & MBC_LED_BLINK_EVT )
  {
    //blink the LED if not connected
    if (modbusState != BLE_STATE_CONNECTED && modbusState != BLE_STATE_CONNECTING){
      LED_TOGGLE();
      if (modbusScanning){
        //blink quickly
        osal_start_timerEx(modbusTaskId, MBC_LED_BLINK_EVT, MBC_DISC_BLINK_PERIOD);
      } else {
        //blink slowly
        osal_start_timerEx(modbusTaskId, MBC_LED_BLINK_EVT, MBC_IDLE_BLINK_PERIOD);
      }
    }
    
    return (events ^ MBC_LED_BLINK_EVT);
  }
  
  // Discard unknown events
  return 0;
}

/*****************************************************************//**
 * @fn      modbusCentral_ProcessOSALMsg
 *
 * @brief   Process an incoming task message.
 *
 * @param   pMsg - message to process
 *
 * @return  none
 */
static void modbusCentral_ProcessOSALMsg( osal_event_hdr_t *pMsg )
{
  switch ( pMsg->event )
  {
    case KEY_CHANGE:
      modbusCentral_HandleKeys( ((keyChange_t *)pMsg)->state, ((keyChange_t *)pMsg)->keys );
      break;

    case GATT_MSG_EVENT:
      modbusCentralProcessGATTMsg( (gattMsgEvent_t *) pMsg );
      break;
  }
}

/*****************************************************************//**
 * @fn      modbusCentral_HandleKeys
 *
 * @brief   Handles all key events for this device.
 *
 * @param   shift - true if in shift/alt.
 * @param   keys - bit field for key events.
 *
 * @return  none
 */
uint8 gStatus;
static void modbusCentral_HandleKeys( uint8 shift, uint8 keys )
{
  (void)shift;  // Intentionally unreferenced parameter
//#ifdef CC2540_RS485
  (void) keys; // Not needed. There is just one key
/*#else
  if ( keys & HAL_KEY_CENTER )
  {
#endif*/
    if ( modbusState != BLE_STATE_CONNECTED ) //discover and connect
    {
      if ( !modbusScanning )
      {
        modbusScanning = TRUE;
        
        LCD_WRITE_STRING( "Discovering...", HAL_LCD_LINE_1 );
        LCD_WRITE_STRING( "", HAL_LCD_LINE_2 );
        
        GAPCentralRole_StartDiscovery( DEFAULT_DISCOVERY_MODE,
                                       DEFAULT_DISCOVERY_ACTIVE_SCAN,
                                       DEFAULT_DISCOVERY_WHITE_LIST );
        
      }
      else
      {
        GAPCentralRole_CancelDiscovery();
      }
    }
    else if ( modbusState == BLE_STATE_CONNECTING ||
              modbusState == BLE_STATE_CONNECTED )
    {
      // disconnect
      modbusState = BLE_STATE_DISCONNECTING;

      gStatus = GAPCentralRole_TerminateLink( modbusConnHandle );
      
      LED_BLINK_START();
      
      LCD_WRITE_STRING( "Disconnecting", HAL_LCD_LINE_1 ); 
    }
/*#ifndef CC2540_RS485    
  }
#endif*/

}

/*****************************************************************//**
 * @fn      modbusCentralProcessGATTMsg
 *
 * @brief   Process GATT messages
 *
 * @return  none
 */
#define SDT_UART_PORT   HAL_UART_PORT_0
static void modbusCentralProcessGATTMsg( gattMsgEvent_t *pMsg )
{
  if ( modbusState != BLE_STATE_CONNECTED )
  {
    // In case a GATT message came after a connection has dropped,
    // ignore the message
    LCD_WRITE_STRING("DISCONNECTED", 1);
    return;
  }
  
  else if ( pMsg->method == ATT_HANDLE_VALUE_NOTI )
  {
    //handle notifications(write them to the UART), 
    //same as the profile change callback in peripheral
    uint8 *data = pMsg->msg.handleValueNoti.value;
    /*if (pMsg->msg.handleValueNoti.len > CHARACTERISTIC_LEN){
      //feil
      LCD_WRITE_STRING_VALUE("Msg len err", pMsg->msg.handleValueNoti.len, 10, 3);
      return;
    } */
    /*HalUARTWrite(SDT_UART_PORT, "beskjed!", 8);*/
    uint8 length = data[0];
    bool final = true;
    if (length & 0x80){
      //not the final notification
      final = false;
      length &= ~0x80;
    }
    if (length > CHARACTERISTIC_LEN-1){
      LCD_WRITE_STRING_VALUE("Too long:", length, 10, 2);
      return;
    } 
    
    if (modbusCurrentLength + length > 256){
      //an error has occured, forget the earlier message (it will probably be invalid)
      modbusCurrentLength = 0;
    }
    if (length > 0)
      osal_memcpy(modbusMessage+modbusCurrentLength, data + 1, length);
    modbusCurrentLength += length;
    
    if (final){
      Modbus_WriteSerial(SDT_UART_PORT, modbusMessage, modbusCurrentLength);
      modbusCurrentLength = 0;
    } 
  }
  
  else if (pMsg->method == ATT_ERROR_RSP) {
    LCD_WRITE_STRING("ERROR!", 1);
    HalUARTWrite(SDT_UART_PORT, "ERROR!", 6);
  }
  
  else if ( modbusDiscState != BLE_DISC_STATE_IDLE )
  {
    modbusGATTDiscoveryEvent( pMsg );
  }
  
  else {
    HalUARTWrite(SDT_UART_PORT, "noke anna...", 12);
  }
  
}

/*****************************************************************//**
 * @fn      modbusCentralEventCB
 *
 * @brief   Central event callback function.
 *
 * @param   pEvent - pointer to event structure
 *
 * @return  none
 */
static void modbusCentralEventCB( gapCentralRoleEvent_t *pEvent )
{
  switch ( pEvent->gap.opcode )
  {
    case GAP_DEVICE_INIT_DONE_EVENT:  
      {
        LCD_WRITE_STRING( "BLE Central", HAL_LCD_LINE_1 );
      }
      break;

    case GAP_DEVICE_INFO_EVENT: //found a device
      {
        // if filtering device discovery results based on service UUID
        if ( DEFAULT_DEV_DISC_BY_SVC_UUID == TRUE )
        {
          if ( modbusFindSvcUuid( SERIALPROFILE_SERV_UUID,
                                     pEvent->deviceInfo.pEvtData,
                                     pEvent->deviceInfo.dataLen ) )
          {
                        
            //FOUND MODBUS!! connect to the device directly
            GAPCentralRole_CancelDiscovery();
            
            uint8* peerAddr = pEvent->deviceInfo.addr;
            uint8 addrType = pEvent->deviceInfo.addrType;
          
            modbusState = BLE_STATE_CONNECTING;
            
            uint8 stat = GAPCentralRole_EstablishLink( DEFAULT_LINK_HIGH_DUTY_CYCLE,
                                          DEFAULT_LINK_WHITE_LIST,
                                          addrType, peerAddr );
            
            if (stat != SUCCESS){
              //procedureInProgress = FALSE;
              LCD_WRITE_STRING( "Not Connecting", HAL_LCD_LINE_1 );
              LCD_WRITE_STRING_VALUE("Status:", stat, 16, HAL_LCD_LINE_2);
            } 
            else {
              LCD_WRITE_STRING( "Connecting", HAL_LCD_LINE_1 );
            }
            
            
          }
        }
      }
      break;
      
    case GAP_DEVICE_DISCOVERY_EVENT:
      {
        // discovery complete
        modbusScanning = FALSE;
        
      }
      break;

    case GAP_LINK_ESTABLISHED_EVENT:
      {
        if ( pEvent->gap.hdr.status == SUCCESS )
        {          
          modbusState = BLE_STATE_CONNECTED;
          modbusConnHandle = pEvent->linkCmpl.connectionHandle;
          //procedureInProgress = TRUE;    

          // If service discovery not performed initiate service discovery
          if ( dataInHandle == 0 || dataOutHandle == 0 )
          {
            osal_start_timerEx( modbusTaskId, START_DISCOVERY_EVT, DEFAULT_SVC_DISCOVERY_DELAY );
          }
                    
          LCD_WRITE_STRING( "Connected", HAL_LCD_LINE_1 );
          
          LED_BLINK_STOP();
          LED_ON();
          //LCD_WRITE_STRING( bdAddr2Str( pEvent->linkCmpl.devAddr ), HAL_LCD_LINE_2 ); 
        
        } 
        else
        {
          modbusState = BLE_STATE_IDLE;
          modbusConnHandle = GAP_CONNHANDLE_INIT;
          //simpleBLERssi = FALSE;
          modbusDiscState = BLE_DISC_STATE_IDLE;
          
          LCD_WRITE_STRING( "Connect Failed", HAL_LCD_LINE_1 );
          LCD_WRITE_STRING_VALUE( "Reason:", pEvent->gap.hdr.status, 10, HAL_LCD_LINE_2 );
        }
      }
      break;

    case GAP_LINK_TERMINATED_EVENT:
      {
        modbusState = BLE_STATE_IDLE;
        modbusConnHandle = GAP_CONNHANDLE_INIT;
        modbusDiscState = BLE_DISC_STATE_IDLE;
        dataInHandle = 0;
        dataOutHandle = 0;
        //procedureInProgress = FALSE;
          
        LCD_WRITE_STRING( "Disconnected", HAL_LCD_LINE_1 );
        LCD_WRITE_STRING_VALUE( "Reason:", pEvent->linkTerminate.reason,
                                10, HAL_LCD_LINE_2 );
        
        LED_BLINK_START();
      }
      break;

    case GAP_LINK_PARAM_UPDATE_EVENT:
      {
        LCD_WRITE_STRING( "Param Update", HAL_LCD_LINE_3 );           
      }
      break;
      
    default:
      break;
  }
}

/******************************************************************//**
 * @fn      modbusCentralStartDiscovery
 *
 * @brief   Start service discovery.
 *
 * @return  none
 */


static void modbusCentralStartDiscovery( void )
{
  //uint8 uuid[ATT_BT_UUID_SIZE] = { LO_UINT16(SERIALPROFILE_SERV_UUID),
  //                                 HI_UINT16(SERIALPROFILE_SERV_UUID) };
  
  uint8 uuid[ATT_UUID_SIZE] = { TI_UUID(SERIALPROFILE_SERV_UUID) };
  
  // Initialize cached handles
  serialSvcStartHdl = serialSvcEndHdl = dataInHandle = dataOutHandle = 0;
  modbusDiscState = BLE_DISC_STATE_SVC;
  
  // Discovery simple BLE service
  GATT_DiscPrimaryServiceByUUID( modbusConnHandle,
                                 uuid,
                                 ATT_UUID_SIZE,
                                 modbusTaskId );
}

uint8 c = 1;
/*****************************************************************//**
 * @fn      modbusGATTDiscoveryEvent
 *
 * @brief   Process GATT discovery event
 *
 * @return  none
 */
__root handleValue128bit_t *values;
static void modbusGATTDiscoveryEvent( gattMsgEvent_t *pMsg )
{
  //attReadByTypeReq_t req;
  HalLcdWriteStringValueValue("DiscEvent:", c++, 10, modbusDiscState, 10, 1);
  if ( modbusDiscState == BLE_DISC_STATE_SVC )
  {
    // Service found, store handles
    if ( pMsg->method == ATT_FIND_BY_TYPE_VALUE_RSP &&
         pMsg->msg.findByTypeValueRsp.numInfo > 0 )
    {
      serialSvcStartHdl = pMsg->msg.findByTypeValueRsp.handlesInfo[0].handle;
      serialSvcEndHdl = pMsg->msg.findByTypeValueRsp.handlesInfo[0].grpEndHandle;
    }
    
    // If procedure complete
    if ( ( pMsg->method == ATT_FIND_BY_TYPE_VALUE_RSP  && 
           pMsg->hdr.status == bleProcedureComplete ) ||
         ( pMsg->method == ATT_ERROR_RSP ) )
    {
      if ( serialSvcStartHdl != 0 )
      {
        modbusDiscState = BLE_DISC_STATE_CHAR;
        
        bStatus_t status = GATT_DiscAllChars(modbusConnHandle, serialSvcStartHdl, serialSvcEndHdl, modbusTaskId);
      }
    }
  }
  else if ( modbusDiscState == BLE_DISC_STATE_CHAR )
  {
    LCD_WRITE_STRING("DATA_IN", 2);
    // Characteristic found, store handle
    if ( (pMsg->method == ATT_READ_BY_TYPE_RSP))
    {    
      if( dataInHandle == 0 || dataOutHandle == 0){
        /*
        modbusDiscState = BLE_DISC_STATE_CHAR;
       
        uint8 p = pMsg->msg.readByTypeRsp.numPairs;
        uint8 length = pMsg->msg.readByTypeRsp.len;
        
        for (uint8 i = length; i <= p*length; i+=length){
          tull = BUILD_UINT16(pMsg->msg.readByTypeRsp.dataList[i-2],
                           pMsg->msg.readByTypeRsp.dataList[i-1]);
          switch (tull)
          {
          case SERIALPROFILE_DATA_IN_UUID:
            dataInHandle = BUILD_UINT16(pMsg->msg.readByTypeRsp.dataList[i-4],
                                        pMsg->msg.readByTypeRsp.dataList[i-3]);
            break;
          case SERIALPROFILE_DATA_OUT_UUID:
            dataOutHandle = BUILD_UINT16(pMsg->msg.readByTypeRsp.dataList[i-4],
                                        pMsg->msg.readByTypeRsp.dataList[i-3]);
            break;
          }
        }*/
        uint8 *hv = pMsg->msg.readByTypeRsp.dataList;
        values = (handleValue128bit_t *)&pMsg->msg.readByTypeRsp.dataList[0];//hv;//(handleValue128bit_t)(pMsg->msg);
        
        uint8 dataLen = pMsg->msg.readByTypeRsp.len;
        uint8 pairs = pMsg->msg.readByTypeRsp.numPairs;
        //uint8 uuid[ATT_UUID_SIZE] = {values + 3};
        for (uint8 i = 0; i < pairs; i++){
          uint8 *b = {hv+3};
          if (osal_memcmp( dataInUUID, values->uuid, ATT_UUID_SIZE) ){
            //found dataIn
            dataInHandle = values->handle;
          }else if (osal_memcmp( dataOutUUID, values->uuid, ATT_UUID_SIZE) ){
            //found dataOut
            dataOutHandle = values->handle;
          } 
          values ++;
        }
      }
      
      if (pMsg->hdr.status == bleProcedureComplete && dataOutHandle != 0){
        //enable notifications
        LCD_WRITE_STRING_VALUE("Data Out hdl:", dataOutHandle, 16, 2);
        
        //enable notifications
        attWriteReq_t req;
        req.handle = dataOutHandle + 1;
        req.len = 2;
        req.value[0] = 1;
        req.value[1] = 0;
        req.sig = 0;
        req.cmd = 0;
        
        bStatus_t status = GATT_WriteCharValue( modbusConnHandle, &req, modbusTaskId );
        
        if (status == SUCCESS)
          notificationsEnabled = true;
        else
          LCD_WRITE_STRING("Notifications disabled", 3);  
        
        modbusDiscState = BLE_DISC_STATE_IDLE;
      }     
    }
    else {
      LCD_WRITE_STRING_VALUE("DI error", pMsg->method, 16, 2);
    }
  }
  else 
  {
    LCD_WRITE_STRING("unknown state",2);
  }
}


/*****************************************************************//**
 * @fn      modbusFindSvcUuid
 *
 * @brief   Find a given UUID in an advertiser's service UUID list.
 *
 * @return  TRUE if service UUID found
 */
static bool modbusFindSvcUuid( uint16 uuid, uint8 *pData, uint8 dataLen )
{
  uint8 adLen;
  uint8 adType;
  uint8 *pEnd;
  
  pEnd = pData + dataLen - 1;
  
  // While end of data not reached
  while ( pData < pEnd )
  {
    // Get length of next AD item
    adLen = *pData++;
    if ( adLen > 0 )
    {
      adType = *pData;
      
      // If AD type is for 16-bit service UUID
      if ( adType == GAP_ADTYPE_128BIT_MORE || adType == GAP_ADTYPE_128BIT_COMPLETE )
      {
        pData++;
        adLen--;
        
        // For each UUID in list
        while ( adLen >= 16 && pData < pEnd )
        {
          // Check for match
          if ( pData[12] == LO_UINT16(uuid) && pData[13] == HI_UINT16(uuid) )
          {
            // Match found
            return TRUE;
          }
          
          // Go to next
          pData += 16;
          adLen -= 16;
        }
        
        // Handle possible erroneous extra byte in UUID list
        if ( adLen == 1 )
        {
          pData++;
        }
      }
      else
      {
        // Go to next item
        pData += adLen;
      }
    }
  }
  
  // Match not found
  return FALSE;
}

/*****************************************************************//**
 * @fn      Modbus_WriteDataIn
 *
 * @brief   Find a given UUID in an advertiser's service UUID list.
 *
 * @param   data - the location of the data to write
 * @param   len - the length of the data
 *
 * @return  The result of GATT_WriteNoRsp, or bleInvalidRange
 */
uint8 first = 0;
bStatus_t Modbus_WriteDataIn(uint8 data[], uint8 len){
  // Do a write
  attWriteReq_t req;
  if (first > 0){
    LCD_WRITE_STRING("NOT FIRST", 1);
    first = 14;
  }
  first++;
  req.handle = dataInHandle;
  req.len = len;
  
  req.sig = 0;
  req.cmd = 1;
  
  if(len > CHARACTERISTIC_LEN)
    return bleInvalidRange;
  
  osal_memcpy(req.value, data, len);
  //for (uint8 i = 0; i < len; i++){
  //  req.value[i] = data[i];
  //}
  return GATT_WriteNoRsp(modbusConnHandle, &req);
}

#if defined CC2540_RS485

HAL_ISR_FUNCTION( buttonISR, P1INT_VECTOR )
{
  HAL_ENTER_ISR();
  //shcedule the HAL_KEY_EVENT
  if (P1IFG & 0x01){ //the button is the interrupt source
    P1IFG= ~0x01; //clear the interrupt flag (only 0 is writable to the interrupt flag register
    osal_start_timerEx (modbusTaskId, MBC_KEY_EVT, MBC_KEY_DEBOUNCE);
  }
  
  P1IFG = 0;
  P1IF = 0;
  
  HAL_EXIT_ISR();
}

#endif
/*********************************************************************
*********************************************************************/
