#include "bcomdef.h"
#include "OSAL.h"
#include "att.h"

#include "hal_lcd.h"
#include "hal_dma.h"
#include "hal_uart.h"
#include "hal_drivers.h"

#include "SerialProfile.h"

#include "modbusDataTransfer.h"
#include "modbusPeripheral.h"
  
/*********************************************************************
 * CONSTANTS
 */

#define MY_DMA_CH_RX            4
#define TIMER1_STOP_CLEAR()     T1CTL = 0x04; T1CNTL = 0x00
#define TIMER1_START()          T1CTL = 0x05
#define TIMER1_RESET()          T1CNTL = 0


/*********************************************************************
 * TYPEDEFS
 */


/*********************************************************************
 * LOCAL VARIABLES
 */

#if HAL_LCD
  uint8 reads = 0;
  uint8 BLEwrites = 1;
#endif
uint8 bytes = 0;
uint8 timer1Reset = 0x00;

uint8 modbusTaskId = 0;

/*********************************************************************
 * LOCAL FUNCTIONS
 */

void sdtUARTcallBack(uint8 port, uint8 event);


/*********************************************************************
 * PUBLIC FUNCTIONS
 */
   
/*********************************************************************
 * @fn      Modbus_initUART
 *
 * @brief   Sets up the UART with the desired settings, as specified
 *          in the header file. Also sets up P0_4 as the DE control 
 *          signal and initializes it to zero
 *
 * @param    none
 *
 * @return   none
 */

void Modbus_initUART(){
#if HAL_LCD
    HalLcdWriteString("UART initializing...", 1);    
#endif
    halUARTCfg_t uartConfig;
    
    //configure UART
    uartConfig.baudRate                 = SDT_UART_BR;
    uartConfig.flowControl              = SDT_UART_FC;
    uartConfig.flowControlThreshold     = SDT_UART_FC_THRESHOLD;
    uartConfig.rx.maxBufSize            = SDT_UART_RX_BUF_SIZE;
    uartConfig.tx.maxBufSize            = SDT_UART_TX_BUF_SIZE;
    uartConfig.idleTimeout              = SDT_UART_IDLE_TIMEOUT;
    uartConfig.intEnable                = SDT_UART_INT_ENABLE;
    uartConfig.callBackFunc             = sdtUARTcallBack; //insert CB here
    uartConfig.configured               = TRUE;
    
    uint8 parity, bit9, d9;
    //even parity
    if (MODBUS_USE_PARITY){
    bit9 = 0x10; //nine bits
    parity = 0x08; //parity bit  
      if (MODBUS_PARITY == EVEN)
        d9 = 0x20; //even parity
      else
        d9 = 0x00; //odd parity 
    }   
    
    //start UART
    HalUARTOpen(SDT_UART_PORT, &uartConfig);
#if HAL_LCD
    HalLcdWriteString("UART ready...", 1);
#endif    
    U0UCR |= parity | bit9 | d9;
    uint8 c; 
    while (HalUARTRead(HAL_UART_PORT_0, &c, 1)){}
    
    //enable uart rx interrupts
    URX0IE = 1;
    
    P0DIR |= 0x30; //pin 4 and 5 are outputs
    P0SEL &= ~0x30; //pin 4 and 5 are GPIO pins
        
   CLEAR_DE(); 
}

/*********************************************************************
 * @fn      Modbus_WriteSerial
 *
 * @brief   Write a message to the serial interface, including control signals
 *
 * @param    uartPort - the port to write
 * @param    data - the data to write
 * @param    length - the length of the data
 *
 * @return   Number of bytes written to the UART buffer
 */
uint8 Modbus_WriteSerial(uint8 uartPort, uint8* data, uint16 length){
  if (length > 0 && length < HAL_UART_DMA_TX_MAX){
    SET_DE();
    return HalUARTWrite(uartPort, data, length);
  }
  return 0;
}



/*********************************************************************
 * @fn      sdtUARTcallBack
 *
 * @brief   The callback function registered with the UART config struct
 *
 * @param    port - the port that had an UART event
 * @param    event - the event type, as defined in hal_uart.h
 *
 * @return   none
 */
void sdtUARTcallBack(uint8 port, uint8 event) {
   
  if(event & HAL_UART_TX_EMPTY){
    while (U0CSR & 0x01); //wait until the UART is finished transmitting (rx is disabled until DE is cleared)
    CLEAR_DE();
  } else if (event & HAL_UART_RX_ABOUT_FULL){
    //reset the timer
    TIMER1_RESET();
    CLEAR_TEST();
  }
}

/*********************************************************************
 * @fn      RS485_InitBoard
 *
 * @brief   Initialize the other board peripherals
 *
 * @param    none
 *
 * @return   none
 */
void RS485_InitBoard(uint8 taskId){
   //configure pins 0 and 1 as GPIO
   P1SEL &= ~0x03;
   //configure pin 0 as an input and pin 1 as output
   P1DIR &= ~0x01;
   P1DIR |= 0x02;
   //enable interrupts on pin 0, disable other pins
   P1IEN = 0x01;
   //enable interrupts on port1
   IEN2 |= 0x10;
   
   modbusTaskId = taskId;
}

/*********************************************************************
 * @fn      RS485_enableTest
 *
 * @brief   Set up test pin
 *
 * @param    none
 *
 * @return   none
 */
void RS485_enableTest(){
  //configure P0_6 as GPIO
  P0SEL &= ~0x40; 
  //configure P0_6 as output
  P0DIR |= 0x40;
  //initialize the state to 0
  P0 &= ~0x40;
  
}

/*********************************************************************
 * @fn      Modbus_initRx
 *
 * @brief   Initializes the serial rx functionality. Setup Timer1
 *          and a open DMA channel to reset the timer
 *
 * @param    none
 *
 * @return   none
 */
void Modbus_initTimer(){
  //initialize timer 1 to be used as the rx-timer
  uint8 prescaler = 0x04; //0b00000100, tick frequency / 8
  uint8 mode = 0x00; //operation suspended
  T1CTL = prescaler | mode;
  
  uint8 t1mode = 0x04; //0b00000100, compare mode
  uint8 im = 0x40; //0b01000000 enable interrupts
  T1CCTL0 = im | t1mode;
  
  //compare value
  T1CC0L = LO_UINT16(MODBUS_RX_TIMEOUT);
  T1CC0H = HI_UINT16(MODBUS_RX_TIMEOUT);
  
  //enable timer1 interrupts
  T1IE = 1;
  
  //setup dma
  halDMADesc_t *ch;

  ch = HAL_DMA_GET_DESC1234( MY_DMA_CH_RX );
  HAL_DMA_SET_SOURCE(ch, timer1Reset);
  HAL_DMA_SET_LEN(ch, 1);
  HAL_DMA_SET_VLEN(ch, HAL_DMA_VLEN_USE_LEN);
  HAL_DMA_SET_DEST(ch, T1CNTL); //writing to T1CNTL clears the timer
  HAL_DMA_SET_WORD_SIZE(ch, HAL_DMA_WORDSIZE_BYTE);
  HAL_DMA_SET_TRIG_MODE(ch, HAL_DMA_TMODE_SINGLE_REPEATED);
  HAL_DMA_SET_TRIG_SRC(ch, HAL_DMA_TRIG_URX0);
  HAL_DMA_SET_SRC_INC(ch, HAL_DMA_SRCINC_0);
  HAL_DMA_SET_DST_INC(ch, HAL_DMA_DSTINC_0);
  HAL_DMA_SET_IRQ(ch,HAL_DMA_IRQMASK_DISABLE);
  HAL_DMA_SET_M8(ch, HAL_DMA_M8_USE_8_BITS);
  HAL_DMA_SET_PRIORITY(ch, HAL_DMA_PRI_HIGH);

  HAL_DMA_CLEAR_IRQ(MY_DMA_CH_RX);
  HAL_DMA_ARM_CH(MY_DMA_CH_RX);
  
}

/*********************************************************************
 * @fn      Modbus_GetFragment
 *
 * @brief   Loads a message fragment of the specified length into the data buffer, if the length is valid
 *
 * @param    data - the location where the data will be stored, will be NULL if the fragment cannot be collected
 * @param    length - the desired length of the fragment
 *
 * @return   none
 */
void Modbus_GetFragment(uint8 *data, uint8 length){
  uint8 avail = Hal_UART_RxBufLen(SDT_UART_PORT);
  uint8 tag = (length | 0x80);
  
  if (avail > length){
    HalUARTRead(SDT_UART_PORT, data+1, length);
    data[0] = tag;
  }
  else {
    data = 0;
  }
}


HAL_ISR_FUNCTION( timer1isr, T1_VECTOR ) { //3.5 bytes of silence, command is finished
  HAL_ENTER_ISR();  
  
  TIMER1_STOP_CLEAR();
  CLEAR_TEST(); 
  osal_set_event(modbusTaskId, MBP_FINAL_FRAGMENT_EVT);
  bytes = 0;
  
  //clear interrupt flags (DISABLE OVERFLOW INTERRUPT!)
  T1STAT = 0;
  T1IE = 0;
  
  HAL_EXIT_ISR();
}

HAL_ISR_FUNCTION( uartRxIsr, URX0_VECTOR ) {
  //start the timer, increment bytes
  
  HAL_ENTER_ISR();
  
  SET_TEST();
  
  TIMER1_RESET();
  TIMER1_START();
  T1IE = 1;
  bytes++;
 
  if (bytes >= CHARACTERISTIC_LEN){
    bytes -= CHARACTERISTIC_LEN;
    //notify the app of the fragment
    osal_set_event(modbusTaskId, MBP_FRAGMENT_READY_EVT);
  }  
  
  HAL_EXIT_ISR();
}
