/**************************************************************************************************
  Filename:       simpleBLEPeripheral.c
  Revised:        $Date: 2010-08-06 08:56:11 -0700 (Fri, 06 Aug 2010) $
  Revision:       $Revision: 23333 $

  Description:    This file contains the Simple BLE Peripheral sample application
                  for use with the CC2540 Bluetooth Low Energy Protocol Stack.

  Copyright 2010 - 2013 Texas Instruments Incorporated. All rights reserved.

  IMPORTANT: Your use of this Software is limited to those specific rights
  granted under the terms of a software license agreement between the user
  who downloaded the software, his/her employer (which must be your employer)
  and Texas Instruments Incorporated (the "License").  You may not use this
  Software unless you agree to abide by the terms of the License. The License
  limits your use, and you acknowledge, that the Software may not be modified,
  copied or distributed unless embedded on a Texas Instruments microcontroller
  or used solely and exclusively in conjunction with a Texas Instruments radio
  frequency transceiver, which is integrated into your product.  Other than for
  the foregoing purpose, you may not use, reproduce, copy, prepare derivative
  works of, modify, distribute, perform, display or sell this Software and/or
  its documentation for any purpose.

  YOU FURTHER ACKNOWLEDGE AND AGREE THAT THE SOFTWARE AND DOCUMENTATION ARE
  PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
  INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY, TITLE,
  NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL
  TEXAS INSTRUMENTS OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER CONTRACT,
  NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR OTHER
  LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
  INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE
  OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT
  OF SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
  (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

  Should you have any questions regarding your right to use this Software,
  contact Texas Instruments Incorporated at www.TI.com.
**************************************************************************************************/

/*********************************************************************
 * INCLUDES
 */

#include "bcomdef.h"
#include "OSAL.h"
#include "OSAL_PwrMgr.h"

#include "OnBoard.h"
#include "hal_adc.h"
#include "hal_led.h"
#include "hal_key.h"
#include "hal_lcd.h"
#include "hal_uart.h"
#include "hal_drivers.h"

#include "gatt.h"

#include "hci.h"

#include "gapgattserver.h"
#include "gattservapp.h"
#include "devinfoservice.h"
#include "SerialProfile.h"

#include "peripheral.h"

#include "gapbondmgr.h"

//#include "simpleBLEPeripheral.h"
#include "modbusPeripheral.h"
#include "modbusDataTransfer.h"

#if defined FEATURE_OAD
  #include "oad.h"
  #include "oad_target.h"
#endif

/*********************************************************************
 * MACROS
 */
uint16 sends = 1;
/*********************************************************************
 * CONSTANTS
 */

   
//key debounce value
#define MBP_KEY_DEBOUNCE                          50

// LED blink period
#define MBP_LED_BLINK_PERIOD                      100
#define MBP_IDLE_BLINK_PERIOD                    1000

// Message fragment timeout in milliseconds, must be larger than the connection interval!
#define MBP_FRAGMENT_TIMEOUT                      10000

// how long to advertise when the button is pushed (in milliseconds)
#define MBP_ADVERTISEMENT_TIMEOUT                 20000

// What is the advertising interval when device is discoverable (units of 625us, 160=100ms)
#define DEFAULT_ADVERTISING_INTERVAL          160

// Limited discoverable mode advertises for 30.72s, and then stops
// General discoverable mode advertises indefinitely

#define DEFAULT_DISCOVERABLE_MODE             GAP_ADTYPE_FLAGS_GENERAL

// Minimum connection interval (units of 1.25ms, 80=100ms) if automatic parameter update request is enabled
#define DEFAULT_DESIRED_MIN_CONN_INTERVAL     8//80

// Maximum connection interval (units of 1.25ms, 800=1000ms) if automatic parameter update request is enabled
#define DEFAULT_DESIRED_MAX_CONN_INTERVAL     80//800

// Slave latency to use if automatic parameter update request is enabled
#define DEFAULT_DESIRED_SLAVE_LATENCY         0

// Supervision timeout value (units of 10ms, 1000=10s) if automatic parameter update request is enabled
#define DEFAULT_DESIRED_CONN_TIMEOUT          1000

// Whether to enable automatic parameter update request when a connection is formed
#define DEFAULT_ENABLE_UPDATE_REQUEST         TRUE

// Connection Pause Peripheral time value (in seconds)
#define DEFAULT_CONN_PAUSE_PERIPHERAL         6

// Company Identifier: Texas Instruments Inc. (13)
#define TI_COMPANY_ID                         0x000D

#define INVALID_CONNHANDLE                    0xFFFF

// Length of bd addr as a string
#define B_ADDR_STR_LEN                        15

/*********************************************************************
 * TYPEDEFS
 */

/*********************************************************************
 * GLOBAL VARIABLES
 */

/*********************************************************************
 * EXTERNAL VARIABLES
 */

/*********************************************************************
 * EXTERNAL FUNCTIONS
 */

/*********************************************************************
 * LOCAL VARIABLES
 */
uint8 modbusMessage[256];
uint8 modbusCurrentLength; 

bool advertising = false;
bool connected = false;
   
static uint8 modbusTaskId;   // Task ID for internal task/event processing

static gaprole_States_t gapProfileState = GAPROLE_INIT;

// GAP - SCAN RSP data (max size = 31 bytes)
static uint8 scanRspData[] =
{
  // complete name
  0x07,   // length of this data
  GAP_ADTYPE_LOCAL_NAME_COMPLETE,
  'M','o','d', 'b', 'u', 's',

  // connection interval range
  0x05,   // length of this data
  GAP_ADTYPE_SLAVE_CONN_INTERVAL_RANGE,
  LO_UINT16( DEFAULT_DESIRED_MIN_CONN_INTERVAL ),   // 10ms
  HI_UINT16( DEFAULT_DESIRED_MIN_CONN_INTERVAL ),
  LO_UINT16( DEFAULT_DESIRED_MAX_CONN_INTERVAL ),   // 100ms
  HI_UINT16( DEFAULT_DESIRED_MAX_CONN_INTERVAL ),

  // Tx power level
  0x02,   // length of this data
  GAP_ADTYPE_POWER_LEVEL,
  0       // 0dBm
};

// GAP - Advertisement data (max size = 31 bytes, though this is
// best kept short to conserve power while advertisting)
static uint8 advertData[] =
{
  // Flags; this sets the device to use limited discoverable
  // mode (advertises for 30 seconds at a time) instead of general
  // discoverable mode (advertises indefinitely)
  0x02,   // length of this data
  GAP_ADTYPE_FLAGS,
  DEFAULT_DISCOVERABLE_MODE | GAP_ADTYPE_FLAGS_BREDR_NOT_SUPPORTED,

  // service UUID, to notify central devices what services are included
  // in this peripheral
  //use 128 but UUID
  0x11,
  GAP_ADTYPE_128BIT_MORE,
  TI_UUID(SERIALPROFILE_SERV_UUID),

};

// GAP GATT Attributes
static uint8 attDeviceName[GAP_DEVICE_NAME_LEN] = "Modbus Peripheral";

/*********************************************************************
 * LOCAL FUNCTIONS
 */
static void modbusPeripheral_ProcessOSALMsg( osal_event_hdr_t *pMsg );
static void peripheralStateNotificationCB( gaprole_States_t newState );
//static void performPeriodicTask( void );
static void serialProfileChangeCB( uint8 paramID );


#if defined( CC2540_MINIDK )
static void simpleBLEPeripheral_HandleKeys( uint8 shift, uint8 keys );
#endif

#if (defined HAL_LCD) && (HAL_LCD == TRUE)
static char *bdAddr2Str ( uint8 *pAddr );
#endif // (defined HAL_LCD) && (HAL_LCD == TRUE)



/*********************************************************************
 * PROFILE CALLBACKS
 */

// GAP Role Callbacks
static gapRolesCBs_t modbusPeripheral_PeripheralCBs =
{
  peripheralStateNotificationCB,  // Profile State Change Callbacks
  NULL                            // When a valid RSSI is read from controller (not used by application)
};

// GAP Bond Manager Callbacks
static gapBondCBs_t modbusPeripheral_BondMgrCBs =
{
  NULL,                     // Passcode callback (not used by application)
  NULL                      // Pairing / Bonding state Callback (not used by application)
};

// Simple GATT Profile Callbacks
static serialProfileCBs_t modbusPeripheral_SerialProfileCBs =
{
  serialProfileChangeCB    // Charactersitic value change callback
};
/*********************************************************************
 * PUBLIC FUNCTIONS
 */

/*********************************************************************
 * @fn      ModbusPeripheral_Init
 *
 * @brief   Initialization function for the Modbus Peripheral App Task.
 *          This is called during initialization and should contain
 *          any application specific initialization (ie. hardware
 *          initialization/setup, table initialization, power up
 *          notificaiton ... ).
 *
 * @param   task_id - the ID assigned by OSAL.  This ID should be
 *                    used to send messages and set timers.
 *
 * @return  none
 */
void ModbusPeripheral_Init( uint8 task_id )
{
  modbusTaskId = task_id;

  // Setup the GAP
  VOID GAP_SetParamValue( TGAP_CONN_PAUSE_PERIPHERAL, DEFAULT_CONN_PAUSE_PERIPHERAL );
  
  // Setup the GAP Peripheral Role Profile
  {
    #if defined CC2540_RS485
      advertising = FALSE;
    #else
      // For other hardware platforms, device starts advertising upon initialization
      advertising = FALSE;
    #endif
    advertising = TRUE;
    connected = FALSE;
    // By setting this to zero, the device will go into the waiting state after
    // being discoverable for 30.72 second, and will not being advertising again
    // until the enabler is set back to TRUE
    uint16 gapRole_AdvertOffTime = 0;

    uint8 enable_update_request = DEFAULT_ENABLE_UPDATE_REQUEST;
    uint16 desired_min_interval = DEFAULT_DESIRED_MIN_CONN_INTERVAL;
    uint16 desired_max_interval = DEFAULT_DESIRED_MAX_CONN_INTERVAL;
    uint16 desired_slave_latency = DEFAULT_DESIRED_SLAVE_LATENCY;
    uint16 desired_conn_timeout = DEFAULT_DESIRED_CONN_TIMEOUT;

    // Set the GAP Role Parameters
    GAPRole_SetParameter( GAPROLE_ADVERT_ENABLED, sizeof( uint8 ), &advertising );
    GAPRole_SetParameter( GAPROLE_ADVERT_OFF_TIME, sizeof( uint16 ), &gapRole_AdvertOffTime );

    GAPRole_SetParameter( GAPROLE_SCAN_RSP_DATA, sizeof ( scanRspData ), scanRspData );
    GAPRole_SetParameter( GAPROLE_ADVERT_DATA, sizeof( advertData ), advertData );

    GAPRole_SetParameter( GAPROLE_PARAM_UPDATE_ENABLE, sizeof( uint8 ), &enable_update_request );
    GAPRole_SetParameter( GAPROLE_MIN_CONN_INTERVAL, sizeof( uint16 ), &desired_min_interval );
    GAPRole_SetParameter( GAPROLE_MAX_CONN_INTERVAL, sizeof( uint16 ), &desired_max_interval );
    GAPRole_SetParameter( GAPROLE_SLAVE_LATENCY, sizeof( uint16 ), &desired_slave_latency );
    GAPRole_SetParameter( GAPROLE_TIMEOUT_MULTIPLIER, sizeof( uint16 ), &desired_conn_timeout );
  }

  // Set the GAP Characteristics
  GGS_SetParameter( GGS_DEVICE_NAME_ATT, GAP_DEVICE_NAME_LEN, attDeviceName );

  // Set advertising interval
  {
    uint16 advInt = DEFAULT_ADVERTISING_INTERVAL;

    GAP_SetParamValue( TGAP_LIM_DISC_ADV_INT_MIN, advInt );
    GAP_SetParamValue( TGAP_LIM_DISC_ADV_INT_MAX, advInt );
    GAP_SetParamValue( TGAP_GEN_DISC_ADV_INT_MIN, advInt );
    GAP_SetParamValue( TGAP_GEN_DISC_ADV_INT_MAX, advInt );
  }

  // Setup the GAP Bond Manager
  {
    uint32 passkey = 0; // passkey "000000"
    uint8 pairMode = GAPBOND_PAIRING_MODE_WAIT_FOR_REQ;
    uint8 mitm = TRUE;
    uint8 ioCap = GAPBOND_IO_CAP_DISPLAY_ONLY;
    uint8 bonding = TRUE;
    GAPBondMgr_SetParameter( GAPBOND_DEFAULT_PASSCODE, sizeof ( uint32 ), &passkey );
    GAPBondMgr_SetParameter( GAPBOND_PAIRING_MODE, sizeof ( uint8 ), &pairMode );
    GAPBondMgr_SetParameter( GAPBOND_MITM_PROTECTION, sizeof ( uint8 ), &mitm );
    GAPBondMgr_SetParameter( GAPBOND_IO_CAPABILITIES, sizeof ( uint8 ), &ioCap );
    GAPBondMgr_SetParameter( GAPBOND_BONDING_ENABLED, sizeof ( uint8 ), &bonding );
  }

  // Initialize GATT attributes
  GGS_AddService( GATT_ALL_SERVICES );            // GAP
  GATTServApp_AddService( GATT_ALL_SERVICES );    // GATT attributes
  DevInfo_AddService();                           // Device Information Service
  SerialProfile_AddService( GATT_ALL_SERVICES );  // Simple GATT Profile
#if defined FEATURE_OAD
  VOID OADTarget_AddService();                    // OAD Profile
#endif

  
#if (defined HAL_LCD) && (HAL_LCD == TRUE)

  HalLcdWriteString( "ModbusPeripheral", HAL_LCD_LINE_1 );

#endif // (defined HAL_LCD) && (HAL_LCD == TRUE)

  // Register callback with SimpleGATTprofile
  VOID SerialProfile_RegisterAppCBs( &modbusPeripheral_SerialProfileCBs );

  // Enable clock divide on halt
  // This reduces active current while radio is active and CC254x MCU
  // is halted
  HCI_EXT_ClkDivOnHaltCmd( HCI_EXT_DISABLE_CLK_DIVIDE_ON_HALT ); //changed from enable
  HCI_EXT_HaltDuringRfCmd( HCI_EXT_HALT_DURING_RF_DISABLE ); //added to keep the processor from shutting down

#if defined ( DC_DC_P0_7 )

  // Enable stack to toggle bypass control on TPS62730 (DC/DC converter)
  HCI_EXT_MapPmIoPortCmd( HCI_EXT_PM_IO_PORT_P0, HCI_EXT_PM_IO_PORT_PIN7 );

#endif // defined ( DC_DC_P0_7 )

  /*
  P1SEL &= ~0x04;
  P1DIR |= 0x04;
  */
//#if defined CC2540_RS485
  RS485_InitBoard(modbusTaskId);
//#endif
  
  // Setup a delayed profile startup
  osal_set_event( modbusTaskId, MBP_START_DEVICE_EVT );
  
}

/*********************************************************************
 * @fn      ModbusPeripheral_ProcessEvent
 *
 * @brief   Modbus Peripheral Application Task event processor.  This function
 *          is called to process all events for the task.  Events
 *          include timers, messages and any other user defined events.
 *
 * @param   task_id  - The OSAL assigned task ID.
 * @param   events - events to process.  This is a bit map and can
 *                   contain more than one event.
 *
 * @return  events not processed
 */
uint16 ModbusPeripheral_ProcessEvent( uint8 task_id, uint16 events )
{

  VOID task_id; // OSAL required parameter that isn't used in this function
  
  if ( events & SYS_EVENT_MSG )
  {
    uint8 *pMsg;

    if ( (pMsg = osal_msg_receive( modbusTaskId )) != NULL )
    {
      modbusPeripheral_ProcessOSALMsg( (osal_event_hdr_t *)pMsg );

      // Release the OSAL message
      VOID osal_msg_deallocate( pMsg );
    }

    // return unprocessed events
    return (events ^ SYS_EVENT_MSG);
  }

  if ( events & MBP_START_DEVICE_EVT )
  {
    // Start the Device
    VOID GAPRole_StartDevice( &modbusPeripheral_PeripheralCBs );

    // Start Bond Manager
    VOID GAPBondMgr_Register( &modbusPeripheral_BondMgrCBs );
    
    Modbus_initUART();
    Modbus_initTimer();
    RS485_enableTest();
    // Set timer for first periodic event
    osal_start_timerEx( modbusTaskId, MBP_LED_BLINK_EVT, MBP_IDLE_BLINK_PERIOD );

    return ( events ^ MBP_START_DEVICE_EVT );
  }

  if (events & MBP_FINAL_FRAGMENT_EVT ){
    //send the final fragments
    uint8 data[CHARACTERISTIC_LEN];
    uint8 tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    
    CLEAR_TEST();
    while (tag >= CHARACTERISTIC_LEN){
      Modbus_GetFragment(data, CHARACTERISTIC_LEN-1);
      SerialProfile_SetParameter(SERIALPROFILE_DATA_OUT, CHARACTERISTIC_LEN, data);
      HalLcdWriteStringValue("Sends", sends++, 10, 2);
      tag -= (CHARACTERISTIC_LEN-1);
      //tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    }
    //tag = Hal_UART_RxBufLen(SDT_UART_PORT);
    HalUARTRead(SDT_UART_PORT, data+1, tag);
    data[0] = tag;
    if (tag != 0){
      SerialProfile_SetParameter(SERIALPROFILE_DATA_OUT, tag+1, data);
      HalLcdWriteStringValue("Sends", sends++, 10, 2);
    }
    return ( events & ~(MBP_FRAGMENT_READY_EVT | MBP_FINAL_FRAGMENT_EVT) );    
  }
  
  if ( events & MBP_FRAGMENT_READY_EVT ) {
    //send a fragment
    uint8 data[CHARACTERISTIC_LEN];
    Modbus_GetFragment(data, CHARACTERISTIC_LEN-1);
    if (data != 0){
      SerialProfile_SetParameter(SERIALPROFILE_DATA_OUT, CHARACTERISTIC_LEN, data);
      HalLcdWriteStringValue("Sends", sends++, 10, 2);
    }      
    return ( events ^ MBP_FRAGMENT_READY_EVT );
  }

#if defined CC2540_RS485
  if ( events & MBP_KEY_EVT )
  {
    if (connected) {
      //do nothing?
      //soft reset?
      SystemResetSoft();
      //GAP_TerminateLinkReq(modbusTaskId, GAP_CONNHANDLE_ALL, 0x16);
    } else {
      //toggle advertising status
      uint8 current_adv_status;
      uint8 new_adv_status;
      
      GAPRole_GetParameter(GAPROLE_ADVERT_ENABLED, &current_adv_status);
      
      if (current_adv_status == FALSE)
        new_adv_status = advertising = TRUE;
      else
        new_adv_status = advertising = FALSE;
      
      GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8), &new_adv_status);
      
      LED_BLINK_START();
    }
    return (events ^ MBP_KEY_EVT);
  }
  
  if ( events  & MBP_LED_BLINK_EVT )
  {
    if ( advertising )
    {
      osal_start_timerEx( modbusTaskId, MBP_LED_BLINK_EVT, MBP_LED_BLINK_PERIOD );
    } else if (connected) {
      //ignore the event
      return (events ^ MBP_LED_BLINK_EVT);
    }else 
    {
      osal_start_timerEx(modbusTaskId, MBP_LED_BLINK_EVT, MBP_IDLE_BLINK_PERIOD );
    }
    LED_TOGGLE();
    return (events ^ MBP_LED_BLINK_EVT);
  }
#endif
  
  if ( events & MBP_FRAGMENT_TIMEOUT_EVT){
    //the message timed out and should be forgotten
    modbusCurrentLength = 0;
    return (events ^ MBP_ADV_OFF_EVT);
  }

  // Discard unknown events
  return 0;
}

/*********************************************************************
 * @fn      modbusPeripheral_ProcessOSALMsg
 *
 * @brief   Process an incoming task message.
 *
 * @param   pMsg - message to process
 *
 * @return  none
 */
static void modbusPeripheral_ProcessOSALMsg( osal_event_hdr_t *pMsg )
{
  switch ( pMsg->event )
  {
  default:
    // do nothing
    break;
  }
}

/*********************************************************************
 * @fn      peripheralStateNotificationCB
 *
 * @brief   Notification from the profile of a state change.
 *
 * @param   newState - new state
 *
 * @return  none
 */
static void peripheralStateNotificationCB( gaprole_States_t newState )
{
#ifdef PLUS_BROADCASTER
  static uint8 first_conn_flag = 0;
#endif // PLUS_BROADCASTER
  advertising = connected = false;
  
  switch ( newState )
  {
    case GAPROLE_STARTED:
      {
        uint8 ownAddress[B_ADDR_LEN];
        uint8 systemId[DEVINFO_SYSTEM_ID_LEN];

        GAPRole_GetParameter(GAPROLE_BD_ADDR, ownAddress);

        // use 6 bytes of device address for 8 bytes of system ID value
        systemId[0] = ownAddress[0];
        systemId[1] = ownAddress[1];
        systemId[2] = ownAddress[2];

        // set middle bytes to zero
        systemId[4] = 0x00;
        systemId[3] = 0x00;

        // shift three bytes up
        systemId[7] = ownAddress[5];
        systemId[6] = ownAddress[4];
        systemId[5] = ownAddress[3];

        DevInfo_SetParameter(DEVINFO_SYSTEM_ID, DEVINFO_SYSTEM_ID_LEN, systemId);

        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          // Display device address
          HalLcdWriteString( bdAddr2Str( ownAddress ),  HAL_LCD_LINE_2 );
          HalLcdWriteString( "Initialized",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
      }
      break;

    case GAPROLE_ADVERTISING:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Advertising",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
          advertising = true;
          gapProfileState = newState;
          
      }
      break;

    case GAPROLE_CONNECTED:
      {        
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Connected",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
          LED_BLINK_STOP();
          LED_ON();
          connected = true;
          advertising = false;
#ifdef PLUS_BROADCASTER
        // Only turn advertising on for this state when we first connect
        // otherwise, when we go from connected_advertising back to this state
        // we will be turning advertising back on.
        if ( first_conn_flag == 0 ) 
        {
          uint8 adv_enabled_status = 1;
          GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8), &adv_enabled_status); // Turn on Advertising
          first_conn_flag = 1;
        }
#endif // PLUS_BROADCASTER
        gapProfileState = newState;
      }
      return;

    case GAPROLE_CONNECTED_ADV:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Connected Advertising",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
          advertising = false;
          connected = false;
          GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8), &advertising);
      }
      break;      
    case GAPROLE_WAITING:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Disconnected",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
      }
      break;

    case GAPROLE_WAITING_AFTER_TIMEOUT:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Timed Out",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
          
#ifdef PLUS_BROADCASTER
        // Reset flag for next connection.
        first_conn_flag = 0;
#endif //#ifdef (PLUS_BROADCASTER)
      }
      break;

    case GAPROLE_ERROR:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "Error",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
      }
      break;

    default:
      {
        #if (defined HAL_LCD) && (HAL_LCD == TRUE)
          HalLcdWriteString( "",  HAL_LCD_LINE_3 );
        #endif // (defined HAL_LCD) && (HAL_LCD == TRUE)
      }
      break;

  }

  gapProfileState = newState;

  LED_BLINK_START();
#if !defined( CC2540_MINIDK )
  VOID gapProfileState;     // added to prevent compiler warning with
                            // "CC2540 Slave" configurations
#endif


}

/*********************************************************************
 * @fn      serialProfileChangeCB
 *
 * @brief   Callback from SerialProfile indicating a value change
 *
 * @param   paramID - parameter ID of the value that was changed.
 *
 * @return  none
 */
static void serialProfileChangeCB( uint8 paramID )
{
  uint8 data[CHARACTERISTIC_LEN];
  uint8 length;
  
  HalLcdWriteString("changeCB entered", 1);
  
  switch( paramID )
  {
    case SERIALPROFILE_DATA_IN:
      {
        //stop pending timeouts
        osal_stop_timerEx(modbusTaskId, MBP_FRAGMENT_TIMEOUT_EVT);
        //get the new value
        SerialProfile_GetParameter( SERIALPROFILE_DATA_IN, &data );
               
        length = data[0];
        bool final = true;
        if (length & 0x80){
          final = false;
          length &= ~0x80;
        }
        
        if (length > CHARACTERISTIC_LEN-1){
          //error, too long message
          return;         
        }
        osal_memcpy(modbusMessage+modbusCurrentLength, data + 1, length);
        modbusCurrentLength += length;
        
        if (final) {
          uint8 l = Modbus_WriteSerial(SDT_UART_PORT, modbusMessage, modbusCurrentLength);
          modbusCurrentLength = 0;
        }   
        else {
          //start the timeout-timer
          osal_start_timerEx(modbusTaskId, MBP_FRAGMENT_TIMEOUT_EVT, MBP_FRAGMENT_TIMEOUT);
        }
      }
      break;
    }
}

HAL_ISR_FUNCTION( buttonISR, P1INT_VECTOR )
{
  HAL_ENTER_ISR();
  //shcedule the HAL_KEY_EVENT
  if (P1IFG & 0x01){ //the button is the interrupt source
    P1IFG= ~0x01; //clear the interrupt flag (only 0 is writable to the interrupt flag register
    osal_start_timerEx (modbusTaskId, MBP_KEY_EVT, MBP_KEY_DEBOUNCE);
  }
  
  P1IFG = 0;
  P1IF = 0;
  
  HAL_EXIT_ISR();
}

#if (defined HAL_LCD) && (HAL_LCD == TRUE)
/*****************************************************************//**
 * @fn      bdAddr2Str
 *
 * @brief   Convert Bluetooth address to string. Only needed when
 *          LCD display is used.
 *
 * @return  none
 */
char *bdAddr2Str( uint8 *pAddr )
{
  uint8       i;
  char        hex[] = "0123456789ABCDEF";
  static char str[B_ADDR_STR_LEN];
  char        *pStr = str;

  *pStr++ = '0';
  *pStr++ = 'x';

  // Start from end of addr
  pAddr += B_ADDR_LEN;

  for ( i = B_ADDR_LEN; i > 0; i-- )
  {
    *pStr++ = hex[*--pAddr >> 4];
    *pStr++ = hex[*pAddr & 0x0F];
  }

  *pStr = 0;

  return str;
}
#endif // (defined HAL_LCD) && (HAL_LCD == TRUE)

/*********************************************************************
*********************************************************************/