/**************************************************************************************************
  Filename:       SerialProfile.c
  Revised:        $Date: 2013-05-06 13:33:47 -0700 (Mon, 06 May 2013) $
  Revision:       $Revision: 34153 $

  Description:    This file contains the Serial profile sample GATT service 
                  profile for use with the Modbus application.

  Copyright 2010 - 2013 Texas Instruments Incorporated. All rights reserved.

  IMPORTANT: Your use of this Software is limited to those specific rights
  granted under the terms of a software license agreement between the user
  who downloaded the software, his/her employer (which must be your employer)
  and Texas Instruments Incorporated (the "License").  You may not use this
  Software unless you agree to abide by the terms of the License. The License
  limits your use, and you acknowledge, that the Software may not be modified,
  copied or distributed unless embedded on a Texas Instruments microcontroller
  or used solely and exclusively in conjunction with a Texas Instruments radio
  frequency transceiver, which is integrated into your product.  Other than for
  the foregoing purpose, you may not use, reproduce, copy, prepare derivative
  works of, modify, distribute, perform, display or sell this Software and/or
  its documentation for any purpose.

  YOU FURTHER ACKNOWLEDGE AND AGREE THAT THE SOFTWARE AND DOCUMENTATION ARE
  PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, 
  INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY, TITLE, 
  NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL
  TEXAS INSTRUMENTS OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER CONTRACT,
  NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR OTHER
  LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
  INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE
  OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT
  OF SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
  (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

  Should you have any questions regarding your right to use this Software,
  contact Texas Instruments Incorporated at www.TI.com. 
**************************************************************************************************/

/*********************************************************************
 * INCLUDES
 */
#include "bcomdef.h"
#include "OSAL.h"
#include "linkdb.h"
#include "att.h"
#include "gatt.h"
#include "gatt_uuid.h"
#include "gattservapp.h"
#include "gapbondmgr.h"

#include "SerialProfile.h"
#include "hal_lcd.h"
   
#include "modbusDataTransfer.h"

/*********************************************************************
 * MACROS
 */

/*********************************************************************
 * CONSTANTS
 */

#define SERVAPP_NUM_ATTR_SUPPORTED        8


/*********************************************************************
 * TYPEDEFS
 */

/*********************************************************************
 * GLOBAL VARIABLES
 *
CONST uint8 serialProfileServUUID[ATT_BT_UUID_SIZE] =
{ 
  LO_UINT16(SERIALPROFILE_SERV_UUID), HI_UINT16(SERIALPROFILE_SERV_UUID)
};
   
CONST uint8 serialProfileDataInUUID[ATT_BT_UUID_SIZE] =
{ 
  LO_UINT16(SERIALPROFILE_DATA_IN_UUID), HI_UINT16(SERIALPROFILE_DATA_IN_UUID)
};

CONST uint8 serialProfileDataOutUUID[ATT_BT_UUID_SIZE] =
{ 
  LO_UINT16(SERIALPROFILE_DATA_OUT_UUID), HI_UINT16(SERIALPROFILE_DATA_OUT_UUID)
};
*/
CONST uint8 serialProfileServUUID[ATT_UUID_SIZE] = 
{
  TI_UUID(SERIALPROFILE_SERV_UUID)
};

CONST uint8 serialProfileDataInUUID[ATT_UUID_SIZE] =
{ 
  TI_UUID(SERIALPROFILE_DATA_IN_UUID)
};

CONST uint8 serialProfileDataOutUUID[ATT_UUID_SIZE] =
{ 
  TI_UUID(SERIALPROFILE_DATA_OUT_UUID)
};
/*********************************************************************
 * EXTERNAL VARIABLES
 */

/*********************************************************************
 * EXTERNAL FUNCTIONS
 */

/*********************************************************************
 * LOCAL VARIABLES
 */

static serialProfileCBs_t *serialProfile_AppCBs = NULL;

/*********************************************************************
 * Profile Attributes - variables
 */

// Simple Profile Service attribute
static CONST gattAttrType_t serialProfileService = { ATT_UUID_SIZE, serialProfileServUUID };


/*****************************************************************
 * DATA TRANSFER SERVICE
 */

//SerialProfile Data_In properties
static uint8 serialProfileDataInProps = GATT_PROP_WRITE;

//SerialProfile Data_In value
static uint8 serialProfileDataInValue[CHARACTERISTIC_LEN] = {0};

//SerialProfile Data_In User Description
static uint8 serialProfileDataInUserDesp[10] = "Data In\0";


//SerialProfile Data_Out properties
static uint8 serialProfileDataOutProps = GATT_PROP_NOTIFY/* | GATT_PROP_READ*/;

//SerialProfile Data_Out value
static uint8 serialProfileDataOutValue[CHARACTERISTIC_LEN] = {0};

//SerialProfile Data_Out User Description
static uint8 serialProfileDataOutUserDesp[10] = "Data Out\0";

//SerialProfile Data_Out configuration
static gattCharCfg_t serialProfileDataOutConfig[GATT_MAX_NUM_CONN];


//SerialProfile Data_Rate?

/*********************************************************************
 * Profile Attributes - Table
 */

static gattAttribute_t serialProfileAttrTbl[SERVAPP_NUM_ATTR_SUPPORTED] = 
{
  // Simple Profile Service
  { 
    { ATT_BT_UUID_SIZE, primaryServiceUUID }, /* type */
    GATT_PERMIT_READ,                         /* permissions */
    0,                                        /* handle */
    (uint8 *)&serialProfileService            /* pValue */
  },

    // Data In Declaration
    { 
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ, 
      0,
      &serialProfileDataInProps 
    },

      // Data In Value
      { 
        { ATT_UUID_SIZE, serialProfileDataInUUID },
        /*GATT_PERMIT_READ | */GATT_PERMIT_WRITE, 
        0, 
        serialProfileDataInValue 
      },

      // Data In User Description
      { 
        { ATT_BT_UUID_SIZE, charUserDescUUID },
        GATT_PERMIT_READ, 
        0, 
        serialProfileDataInUserDesp 
      },      

    // Data Out Declaration
    { 
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ, 
      0,
      &serialProfileDataOutProps 
    },

      // Data Out Value
      { 
        { ATT_UUID_SIZE, serialProfileDataOutUUID },
        GATT_PERMIT_READ, 
        0, 
        serialProfileDataOutValue 
      },
      
      // Data Out configuration
      { 
        { ATT_BT_UUID_SIZE, clientCharCfgUUID },
        GATT_PERMIT_READ | GATT_PERMIT_WRITE, 
        0, 
        (uint8 *)serialProfileDataOutConfig 
      },
      
      // Data Out Description
      { 
        { ATT_BT_UUID_SIZE, charUserDescUUID },
        GATT_PERMIT_READ, 
        0, 
        serialProfileDataOutUserDesp 
      },
};


/*********************************************************************
 * LOCAL FUNCTIONS
 */
static uint8 serialProfile_ReadAttrCB( uint16 connHandle, gattAttribute_t *pAttr, 
                            uint8 *pValue, uint8 *pLen, uint16 offset, uint8 maxLen );
static bStatus_t serialProfile_WriteAttrCB( uint16 connHandle, gattAttribute_t *pAttr,
                                 uint8 *pValue, uint8 len, uint16 offset );

static void serialProfile_HandleConnStatusCB( uint16 connHandle, uint8 changeType );


/*********************************************************************
 * PROFILE CALLBACKS
 */
// Simple Profile Service Callbacks
CONST gattServiceCBs_t simpleProfileCBs =
{
  serialProfile_ReadAttrCB,  // Read callback function pointer
  serialProfile_WriteAttrCB, // Write callback function pointer
  NULL                       // Authorization callback function pointer
};

/*********************************************************************
 * PUBLIC FUNCTIONS
 */

/*********************************************************************
 * @fn      SerialProfile_AddService
 *
 * @brief   Initializes the Serial Profile service by registering
 *          GATT attributes with the GATT server.
 *
 * @param   services - services to add. This is a bit map and can
 *                     contain more than one service.
 *
 * @return  Success or Failure
 */
bStatus_t SerialProfile_AddService( uint32 services )
{
  uint8 status = SUCCESS;

  // Initialize Client Characteristic Configuration attributes
  GATTServApp_InitCharCfg( INVALID_CONNHANDLE, serialProfileDataOutConfig );

  // Register with Link DB to receive link status change callback
  VOID linkDB_Register( serialProfile_HandleConnStatusCB );  
  
  if ( services & SERIALPROFILE_SERVICE )
  {
    // Register GATT attribute list and CBs with GATT Server App
    status = GATTServApp_RegisterService( serialProfileAttrTbl, 
                                          GATT_NUM_ATTRS( serialProfileAttrTbl ),
                                          &simpleProfileCBs );
  }

  return ( status );
}


/*********************************************************************
 * @fn      SerialProfile_RegisterAppCBs
 *
 * @brief   Registers the application callback function. Only call 
 *          this function once.
 *
 * @param   callbacks - pointer to application callbacks.
 *
 * @return  SUCCESS or bleAlreadyInRequestedMode
 */
bStatus_t SerialProfile_RegisterAppCBs( serialProfileCBs_t *appCallbacks )
{
  if ( appCallbacks )
  {
    serialProfile_AppCBs = appCallbacks;
    
    return ( SUCCESS );
  }
  else
  {
    return ( bleAlreadyInRequestedMode );
  }
}
  

/*********************************************************************
 * @fn      SerialProfile_SetParameter
 *
 * @brief   Set a Simple Profile parameter.
 *
 * @param   param - Profile parameter ID
 * @param   len - length of data to right
 * @param   value - pointer to data to write.  This is dependent on
 *          the parameter ID and WILL be cast to the appropriate 
 *          data type (example: data type of uint16 will be cast to 
 *          uint16 pointer).
 *
 * @return  bStatus_t
 */
bStatus_t SerialProfile_SetParameter( uint8 param, uint8 len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
    /*
  case SERIALPROFILE_DATA_IN:
      if (len == MODBUS_MAX_LEN )
      {
        VOID osal_memcpy(serialProfileDataInValue, value, MODBUS_MAX_LEN);
      }
      else
      {
        ret = bleInvalidRange;
      }
      break;
    */  
  case SERIALPROFILE_DATA_OUT:
    if (len <= CHARACTERISTIC_LEN)
    {
      VOID osal_memcpy(serialProfileDataOutValue, value, len);
      // See if Notification has been enabled
        ret = GATTServApp_ProcessCharCfg( serialProfileDataOutConfig, serialProfileDataOutValue, FALSE,
                                    serialProfileAttrTbl, GATT_NUM_ATTRS( serialProfileAttrTbl ),
                                    INVALID_TASK_ID );
#if HAL_LCD  
      if (ret != SUCCESS )
          HalLcdWriteString("notification failure", 2);
#endif
    }
    else
    {
      ret = bleInvalidRange;
    }
    break;
    
    default:
      ret = INVALIDPARAMETER;
      break;
  }
  
  return ( ret );
}

/*********************************************************************
 * @fn      SerialProfile_GetParameter
 *
 * @brief   Get a Simple Profile parameter.
 *
 * @param   param - Profile parameter ID
 * @param   value - pointer to data to put.  This is dependent on
 *          the parameter ID and WILL be cast to the appropriate 
 *          data type (example: data type of uint16 will be cast to 
 *          uint16 pointer).
 *
 * @return  bStatus_t
 */
bStatus_t SerialProfile_GetParameter( uint8 param, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {

  case SERIALPROFILE_DATA_IN:
      VOID osal_memcpy(value, serialProfileDataInValue, CHARACTERISTIC_LEN);
      break;
  case SERIALPROFILE_DATA_OUT:
      VOID osal_memcpy(value, serialProfileDataOutValue, CHARACTERISTIC_LEN);
      break;
    
      
    default:
      ret = INVALIDPARAMETER;
      break;
  }
  
  return ( ret );
}

/*********************************************************************
 * @fn          serialProfile_ReadAttrCB
 *
 * @brief       Read an attribute.
 *
 * @param       connHandle - connection message was received on
 * @param       pAttr - pointer to attribute
 * @param       pValue - pointer to data to be read
 * @param       pLen - length of data to be read
 * @param       offset - offset of the first octet to be read
 * @param       maxLen - maximum length of data to be read
 *
 * @return      Success or Failure
 */
static uint8 serialProfile_ReadAttrCB( uint16 connHandle, gattAttribute_t *pAttr, 
                            uint8 *pValue, uint8 *pLen, uint16 offset, uint8 maxLen )
{
  bStatus_t status = SUCCESS;

  // If attribute permissions require authorization to read, return error
  if ( gattPermitAuthorRead( pAttr->permissions ) )
  {
    // Insufficient authorization
    return ( ATT_ERR_INSUFFICIENT_AUTHOR );
  }
  
  // Make sure it's not a blob operation (no attributes in the profile are long)
  if ( offset > 0 )
  {
    return ( ATT_ERR_ATTR_NOT_LONG );
  }
 
  if ( pAttr->type.len == ATT_BT_UUID_SIZE )
  {
    // 16-bit UUID
    uint16 uuid = BUILD_UINT16( pAttr->type.uuid[0], pAttr->type.uuid[1]);
    switch ( uuid )
    {
      // No need for "GATT_SERVICE_UUID" or "GATT_CLIENT_CHAR_CFG_UUID" cases;
      // gattserverapp handles those reads
             
      case SERIALPROFILE_DATA_OUT_UUID:
        *pLen = CHARACTERISTIC_LEN;
        VOID osal_memcpy(pValue, pAttr->pValue, CHARACTERISTIC_LEN);
        break;
        
      default:
        // Should never get here! 
        *pLen = 0;
        status = ATT_ERR_ATTR_NOT_FOUND;
        break;
    }
  }
  else
  {
    // 128-bit UUID
    uint8 uuid[ATT_UUID_SIZE]; 
    osal_memcpy(uuid, pAttr->type.uuid, ATT_UUID_SIZE);
    if (osal_memcmp(uuid, serialProfileDataOutUUID, ATT_UUID_SIZE))
    {
      *pLen = CHARACTERISTIC_LEN;
      VOID osal_memcpy( pValue, pAttr->pValue, CHARACTERISTIC_LEN );
    }
    else {
      *pLen = 0;
      status = ATT_ERR_ATTR_NOT_FOUND;
    }
  }

  return ( status );
}

/*********************************************************************
 * @fn      serialProfile_WriteAttrCB
 *
 * @brief   Validate attribute data prior to a write operation
 *
 * @param   connHandle - connection message was received on
 * @param   pAttr - pointer to attribute
 * @param   pValue - pointer to data to be written
 * @param   len - length of data
 * @param   offset - offset of the first octet to be written
 *
 * @return  Success or Failure
 */
static bStatus_t serialProfile_WriteAttrCB( uint16 connHandle, gattAttribute_t *pAttr,
                                 uint8 *pValue, uint8 len, uint16 offset )
{
  bStatus_t status = SUCCESS;
  uint8 notifyApp = 0xFF;
  
  // If attribute permissions require authorization to write, return error
  if ( gattPermitAuthorWrite( pAttr->permissions ) )
  {
    // Insufficient authorization
    return ( ATT_ERR_INSUFFICIENT_AUTHOR );
  }
  
  if ( pAttr->type.len == ATT_BT_UUID_SIZE )
  {
    // 16-bit UUID
    uint16 uuid = BUILD_UINT16( pAttr->type.uuid[0], pAttr->type.uuid[1]);
    switch ( uuid )
    {
      case SERIALPROFILE_DATA_IN_UUID:

        // Validate the value
        // Make sure it's not a blob oper
        if ( offset == 0 )
        {
          if ( len > CHARACTERISTIC_LEN )
          {
            status = ATT_ERR_INVALID_VALUE_SIZE;
          }
        }
        else
        {
          status = ATT_ERR_ATTR_NOT_LONG;
        }
        
        //Write the value
        if ( status == SUCCESS )
        {
          VOID osal_memcpy(pAttr->pValue, pValue, CHARACTERISTIC_LEN);
          
          notifyApp = SERIALPROFILE_DATA_IN;
        }
             
        break;

      case GATT_CLIENT_CHAR_CFG_UUID:
        HalLcdWriteString("Enabled notifications", 2);
        status = GATTServApp_ProcessCCCWriteReq( connHandle, pAttr, pValue, len,
                                                 offset, GATT_CLIENT_CFG_NOTIFY );
        break;
        
      default:
        // Should never get here! (characteristics 2 and 4 do not have write permissions)
        status = ATT_ERR_ATTR_NOT_FOUND;
        break;
    }
  }
  else
  {
    // 128-bit UUID
    uint8 uuid[ATT_UUID_SIZE];
    osal_memcpy(uuid, pAttr->type.uuid, ATT_UUID_SIZE);
    
    if (osal_memcmp(uuid, serialProfileDataInUUID, ATT_UUID_SIZE)){
      if ( offset == 0 )
      {
          if ( len > CHARACTERISTIC_LEN )
          {
            status = ATT_ERR_INVALID_VALUE_SIZE;
          }
               
        //Write the value
        if ( status == SUCCESS )
        {
          VOID osal_memcpy(pAttr->pValue, pValue, CHARACTERISTIC_LEN);
          
          notifyApp = SERIALPROFILE_DATA_IN;
        }
      } else {
          status = ATT_ERR_ATTR_NOT_LONG;
      }
    } else {
          status = ATT_ERR_ATTR_NOT_FOUND;
      }
  }

  // If a charactersitic value changed then callback function to notify application of change
  if ( (notifyApp != 0xFF ) && serialProfile_AppCBs && serialProfile_AppCBs->pfnSimpleProfileChange )
  {
    serialProfile_AppCBs->pfnSimpleProfileChange( notifyApp );  
  }
  
  return ( status );
}

/*********************************************************************
 * @fn          serialProfile_HandleConnStatusCB
 *
 * @brief       Simple Profile link status change handler function.
 *
 * @param       connHandle - connection handle
 * @param       changeType - type of change
 *
 * @return      none
 */
static void serialProfile_HandleConnStatusCB( uint16 connHandle, uint8 changeType )
{ 
  // Make sure this is not loopback connection
  if ( connHandle != LOOPBACK_CONNHANDLE )
  {
    // Reset Client Char Config if connection has dropped
    if ( ( changeType == LINKDB_STATUS_UPDATE_REMOVED )      ||
         ( ( changeType == LINKDB_STATUS_UPDATE_STATEFLAGS ) && 
           ( !linkDB_Up( connHandle ) ) ) )
    { 
      GATTServApp_InitCharCfg( connHandle, serialProfileDataOutConfig );
    }
  }
}


/*********************************************************************
*********************************************************************/
